import glob
import os

import numpy as np
import pytest

import micropolarray as ml

# import micropolarray as ml
from micropolarray.processing.demodulation import Malus


@pytest.fixture(autouse=True)
def generate_dummy_data():
    """Dummy data factory"""

    def _make_dummy_data(dimension):
        dummydata = np.zeros(shape=(dimension, dimension))
        dummydata[0::2, 0::2] = 1
        dummydata[0::2, 1::2] = 2
        dummydata[1::2, 0::2] = 3
        dummydata[1::2, 1::2] = 4
        return dummydata

    return _make_dummy_data


def generate_polarized_data(
    shape, S, angle_rad=0, t=1, eff=1, angles_list=[0, 45, -45, 90]
):
    single_pol_shape = (int(shape[0] / 2), int(shape[0] / 2))
    ones = np.ones(shape=single_pol_shape)
    angles = np.array([np.deg2rad(angle) for angle in angles_list])
    subimages = np.array(
        [ones * S * Malus(angle_rad, t, eff, angle) for angle in angles]
    )
    return ml.merge_polarizations(subimages)


def generate_polarimetric_measurements(tmp_path, input_signal, t, eff, shape):
    shape = (shape, shape)
    # try demodulation
    angles = np.array([np.deg2rad(angle) for angle in [0, 45, -45, 90]])
    output_dir = tmp_path / "computed_matrix"
    output_str = str(output_dir)

    polarizations = np.arange(-45, 91, 15)
    pols_rad = np.deg2rad(polarizations)

    ones = np.ones(shape=shape)
    for pol, pol_rad in zip(polarizations, pols_rad):
        result_image = ml.MicropolImage(
            generate_polarized_data(
                shape=shape,
                S=input_signal,
                angle_rad=pol_rad,
                t=t,
                eff=eff,
            )
        )
        result_image.save_as_fits(tmp_path / f"pol_{int(pol)}.fits")
    if False:  # check that fit will be ok
        for angle in angles:
            pars, pcov = curve_fit(
                Malus,
                pols_rad,
                np.array([Malus(pol, t, eff, angle) for pol in pols_rad]),
            )
            print(f"t = {pars[0]}")
            print(f"eff = {pars[1] }")
            print(f"phi = {np.rad2deg(pars[2]) }")

    # read the files
    filenames = sorted(
        glob.glob(str(tmp_path / "pol*.fits")),
        key=lambda x: int(x.split(os.path.sep)[-1][4:].strip(".fits")),
    )



import os
import sys
import time
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import constants
from scipy.optimize import curve_fit

from micropolarray.cameras import PolarCam
from micropolarray.processing.demosaic import (
    merge_polarizations,
    split_polarizations,
)


# timer decorator
def timer(func):
    """Use this to time function execution

    Args:
        func (function): function of which to measure execution time
    """

    def wrapper(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(
            f"Function {func.__name__} took {round(end - start, 4)} s to run"
        )
        return result

    return wrapper


def _make_abs_and_create_dir(filename: str):
    path = Path(filename)

    if not path.is_absolute():  # suppose it is in cwd
        path = path.joinpath(Path().cwd(), path)

    if path.suffix:
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
    else:
        if not path.exists():
            path.mkdir()
    return str(path)


def sigma_DN(pix_DN):
    gain = 6.93
    sigma_DN = np.sqrt(gain * pix_DN) / gain
    return sigma_DN


def fix_data(data: np.array, min, max):
    if not (min and max):
        return data
    fixed_data = data.copy()
    fixed_data = np.where(fixed_data > min, fixed_data, min)
    fixed_data = np.where(fixed_data < max, fixed_data, max)
    return fixed_data


def mean_minus_std(data: np.array, stds_n: int = 1) -> float:
    """Returns the value at the mean - standard deviation for the input data

    Args:
        data (np.array): input data
        stds_n (int, optional): number of standard deviations. Defaults to 1.

    Returns:
        float: mean value - n*stdevs
    """
    return np.mean(data) - stds_n * np.std(data)


def mean_plus_std(data: np.array, stds_n: int = 1) -> float:
    """Returns the value at the mean + standard deviation for the input data

    Args:
        data (np.array): input data
        stds_n (int, optional): number of standard deviations. Defaults to 1.

    Returns:
        float: mean value + n*stdevs
    """
    return np.mean(data) + stds_n * np.std(data)


def median_minus_std(data: np.array, stds_n: int = 1) -> float:
    """Returns the value at the median - median deviation for the input data

    Args:
        data (np.array): input data
        stds_n (int, optional): number of standard deviations. Defaults to 1.

    Returns:
        float: median value - n*mediandevs
    """
    median = np.median(data)
    median_std = np.median(np.abs(data - median))
    return median - stds_n * median_std


def median_plus_std(data: np.array, stds_n: int = 1) -> float:
    """Returns the value at the median + median deviation for the input data

    Args:
        data (np.array): input data
        stds_n (int, optional): number of standard deviations. Defaults to 1.

    Returns:
        float: median value + n*mediandevs
    """
    median = np.median(data)
    median_std = np.median(np.abs(data - median))
    return median + stds_n * median_std


def normalize2pi(angles_list):
    if type(angles_list) is not list:
        angles_list = [
            angles_list,
        ]
    for i, angle in enumerate(angles_list):
        while angle > 90:
            angle -= 180
        while angle <= -90:
            angle += 180
        angles_list[i] = angle

    return angles_list


def align_keywords_and_data(header, data, sun_center, platescale, binning=1):
    """Fixes antarticor keywords and data to reflect each other.

    Args:
        header (dict): fits file header
        data (ndarray): data as np array
        platescale (float): plate scale in arcsec/pixel
        binning (int, optional): binning applied to image. Defaults to 1 (no binning).

    Returns:
        header, data: new fixed header and data
    """

    single_pol_images = split_polarizations(data)
    # data = np.rot90(data, k=-1)
    # data = np.flip(data, axis=0)
    for i in range(4):
        single_pol_images[i] = np.rot90(single_pol_images[i], k=-1)
        single_pol_images[i] = np.flip(single_pol_images[i], axis=0)
    data = merge_polarizations(single_pol_images)

    header["NAXIS1"] = data.shape[0]
    header["NAXIS2"] = data.shape[1]
    height = header["NAXIS1"]
    width = header["NAXIS2"]
    rotation_angle = -9  # degrees
    if binning > 1:
        platescale = platescale * binning

    header["DATE-OBS"] = header["DATE-OBS"] + "T" + header["TIME-OBS"]

    header["WCSNAME"] = "helioprojective-cartesian"
    # header["DSUN_OBS"] = 1.495978707e11
    header["CTYPE1"] = "HPLN-TAN"
    header["CTYPE2"] = "HPLT-TAN"
    header["CDELT1"] = platescale
    header["CDELT2"] = platescale
    header["CUNIT1"] = "arcsec"
    header["CUNIT2"] = "arcsec"
    header["CRVAL1"] = 0
    header["CRVAL2"] = 0
    header["CROTA2"] = rotation_angle

    y, x = sun_center
    # if year == 2021:
    #    y, x, _ = PolarCam().occulter_pos_2021
    # elif year == 2022:
    #    y, x, _ = PolarCam().occulter_pos_last
    relative_y = y / 1952
    relative_x = x / 1952
    sun_x = int(width * relative_x)
    sun_y = int(height * relative_y)

    # one changes because of rotation, the other because of jhelioviewer representation
    header["CRPIX1"] = height - sun_y  # y, checked
    header["CRPIX2"] = width - sun_x  # x, checked

    return header, data


def get_Bsun_units(
    diffuser_I: float,
    texp_image: float = 1.0,
    texp_diffuser: float = 1.0,
) -> float:
    """Returns the conversion unit for expressing brightness in units of sun brightness. Usage is
    data [units of B_sun] = data[DN] * get_Bsun_units(mean_Bsun_brightness, texp_image, texp_diffuser)

    Args:
        mean_sun_brightness (float): diffuser mean in DN.
        texp_image (float, optional): image exposure time. Defaults to 1.0.
        texp_diffuser (float, optional): diffuser exposure time. Defaults to 1.0.

    Returns:
        float: Bsun units conversion factor
    """
    diffusion_solid_angle = 1.083 * 1.0e-5
    diffuser_transmittancy = 0.28
    Bsun_unit = (
        diffusion_solid_angle
        * diffuser_transmittancy
        * texp_diffuser
        / texp_image
    )
    Bsun_unit = (
        (1.0 / texp_image)
        * diffuser_transmittancy
        * diffusion_solid_angle
        / (diffuser_I / texp_diffuser)
    )

    return Bsun_unit


def get_malus_normalization(four_peaks_images, show_hist=False):
    S_max = np.zeros_like(
        four_peaks_images[0]
    )  # tk_sum = tk_0 + tk_45 + tk_90 + tk_45
    S_max = 0.5 * np.sum(four_peaks_images, axis=0)
    # Normalizing S, has a spike of which maximum is taken
    bins = 1000
    histo = np.histogram(S_max, bins=bins)
    maxvalue = np.max(histo[0])
    index = np.where(histo[0] == maxvalue)[0][0]
    normalizing_S = (
        histo[1][index] + histo[1][index + 1] + histo[1][index - 1]
    ) / 3
    # normalizing_S = np.max(S_max) # old

    # ----------------------------------------------
    # fit gaussian to S for normalization signal
    def gauss(x, norm, x_0, sigma):
        return norm * np.exp(-((x - x_0) ** 2) / (4 * sigma**2))

    hist_roi = 10  # bins around max value
    xvalues = np.array(histo[1])[index - hist_roi : index + hist_roi]
    yvalues = np.array(histo[0])[index - hist_roi : index + hist_roi]
    yvalues_sum = np.sum(yvalues)
    yvalues = yvalues / yvalues_sum
    xvalues = np.array(
        [value + (xvalues[1] - xvalues[0]) / 2 for value in xvalues]
    )  # shift each bin to center
    prediction = [
        yvalues[int(len(yvalues) / 2)],  # normalization
        xvalues[int(len(xvalues) / 2)],  # center
        xvalues[int(len(xvalues) / 2) + int(hist_roi / 2)]
        - xvalues[int(len(xvalues) / 2)],  # sigma
    ]
    params, cov = curve_fit(
        gauss,
        xvalues,
        yvalues,
        prediction,
    )
    normalizing_S = params[1] + 4 * params[2]  # center of gaussian + 2sigma
    # 3sigma -> P = 2.7e-3 outliers
    # 4sigma -> P = 6.3e-5 outliers
    # ----------------------------------------------
    if show_hist:
        index = 5
        fig, ax = plt.subplots(figsize=(9, 9))
        ax.stairs(histo[0], histo[1], label=f"S, max = {np.max(S_max)}")
        ax.axvline(normalizing_S, color="red", label="normalizing_S")
        ax.plot(
            xvalues,
            gauss(xvalues, params[0] * yvalues_sum, params[1], params[2]),
            label="Fitted curve for normalizing S",
        )
        ax.set_title(f"Normalizing S (t_0 + t_45 + t_90 + t_135)")
        ax.set_xlabel("S [DN]")
        ax.set_ylabel("Counts")
        ax.legend()
        plt.show()
        sys.exit()

    return normalizing_S



import sys
from itertools import product

import numpy as np
from numba import njit

from ..micropol_image import MicropolImage
from ..utils import timer
from .demodulation import Demodulator
from .demosaic import split_polarizations


class MicropolImageError:
    def __init__(
        self, image: MicropolImage, image_error: np.ndarray, demodulator: Demodulator
    ) -> None:
        self.sigma_S = get_error_on_Stokes(
            image_error=image_error, demodulator=demodulator
        )
        self.sigma_pB = get_error_on_pB(image.Stokes_vec, self.sigma_S)
        self.sigma_DoLP = get_error_on_DoLP(image.Stokes_vec, self.sigma_S)
        self.sigma_AoLP = get_error_on_AoLP(image.Stokes_vec, self.sigma_S)


def get_error_on_Stokes(
    image_error: np.ndarray, demodulator: Demodulator
) -> np.ndarray:
    """Returns the error on the image, propagated through the demodulation matrix. If M[i, j] is the demodulation matrix, sigma_I[k] are the four pixel values in a superpixel, and S[i, j] is the Stokes vector, returns the matrix product
    sqrt(M^2 @ I^2)

    Args:
        image_error (np.ndarray): array containing the pixel by pixel error to propagate.
        demodulator (Demodulator): demodulator containing the demodulation matrix.

    Returns:
        np.ndarray: errors of the computed Stokes vector as a [3, y, x] array.
    """
    mij_square = np.multiply(demodulator.mij, demodulator.mij)

    single_pol_subimages = split_polarizations(image_error)
    pixel_poisson_variance = np.expand_dims(
        np.multiply(single_pol_subimages, single_pol_subimages), axis=0
    )

    # S_variance = mij * sigma_image
    S_variance = np.matmul(
        mij_square,
        pixel_poisson_variance,
        axes=[(-4, -3), (-3, -4), (-4, -3)],
    )[:, 0]

    return np.sqrt(S_variance)


def get_error_on_pB(S: np.ndarray, sigma_S: np.ndarray) -> np.ndarray:
    I, Q, U = S
    sigma_I, sigma_Q, sigma_U = sigma_S

    pb_var = (Q * Q * sigma_Q * sigma_Q + U * U * sigma_U * sigma_U) / (Q * Q + U * U)

    return np.sqrt(pb_var)


def get_error_on_DoLP(S: np.ndarray, sigma_S: np.ndarray) -> np.ndarray:
    I, Q, U = S
    sigma_I, sigma_Q, sigma_U = sigma_S

    pB = np.sqrt(Q * Q + U * U)

    dolp_var = (Q * Q * sigma_Q * sigma_Q + U * U * sigma_U * sigma_U) / (
        (I * pB) * (I * pB)
    ) + (pB / (I * I)) * (pB / (I * I)) * sigma_I * sigma_I

    return np.sqrt(dolp_var)


def get_error_on_AoLP(S: np.ndarray, sigma_S: np.ndarray) -> np.ndarray:
    I, Q, U = S
    sigma_I, sigma_Q, sigma_U = sigma_S

    aolp_var = (sigma_U * sigma_U + sigma_Q * sigma_Q * U * U / (Q * Q)) / (
        4 * Q * Q * (1 + (U * U / (Q * Q))) * (1 + (U * U / (Q * Q)))
    )

    return np.sqrt(aolp_var)



import logging

from micropolarray.cameras import Antarticor, Kasi, PolarCam
from micropolarray.image import Image
from micropolarray.micropol_image import MicropolImage
from micropolarray.micropol_image import MicropolImage as MicroPolarizerArrayImage
from micropolarray.micropol_image import MicropolImage as PolarcamImage
from micropolarray.micropol_image import set_default_angles
from micropolarray.processing.chen_wan_liang_calibration import (
    _ifov_jitcorrect,
    chen_wan_liang_calibration,
)
from micropolarray.processing.congrid import congrid
from micropolarray.processing.convert import (
    average_rawfiles_to_fits,
    convert_rawfile_to_fits,
    convert_set,
    merge_rawfiles_to_fits,
)
from micropolarray.processing.demodulation import (
    Demodulator,
    calculate_demodulation_tensor,
)
from micropolarray.processing.demodulation_errors import MicropolImageError
from micropolarray.processing.demosaic import (
    demosaic,
    merge_polarizations,
    split_polarizations,
)
from micropolarray.processing.hotpix_cleaning import get_hot_pixels
from micropolarray.processing.nrgf import (
    find_occulter_hough,
    find_occulter_position,
    nrgf,
    roi_from_polar,
)
from micropolarray.processing.rebin import trim_to_match_binning
from micropolarray.utils import (
    align_keywords_and_data,
    get_Bsun_units,
    get_malus_normalization,
    mean_minus_std,
    mean_plus_std,
    median_minus_std,
    median_plus_std,
    normalize2pi,
    sigma_DN,
)

logging.basicConfig(
    level=logging.INFO, format="%(levelname)s - %(asctime)s - %(message)s"
)  # tempo, livello, messaggio. livello è warning, debug, info, error, critical

__all__ = []  # Imported modules when "from microppolarray_lib import *" is called



import datetime
import multiprocessing as mp
import os
import sys
from logging import critical, info
from pathlib import Path

import numpy as np
import pytz
import tqdm
from astropy.io import fits


def three_bytes_to_two_ints(filecontent):
    """Needed for parallelization, this will be run by each thread for a slice of the original array.

    Returns:
        np.array: array of saved data
    """
    newarr = np.zeros(shape=int(len(filecontent) * 2 / 3))
    position = 0
    for i in range(0, len(filecontent), 3):
        binsum = "".join([bin(j)[2:].zfill(8) for j in filecontent[i : i + 3]])
        newarr[position] = int(binsum[0:12], 2)
        newarr[position + 1] = int(binsum[16:24] + binsum[12:16], 2)
        position += 2
    return newarr


def nparr_from_binary(filename):
    """Converts a PolarCam binary file into a numpy array. Bytes are saved like this

    - 24 bit (3 bytes)
        1             |   3                |     2
        111111111111       | 1111               | 11111111
    - 2 numbers
        First number 12bit | Second number (little endian) 8+4=12 bit

    Args:
        filename (str): name of the file to be converted

    Raises:
        ValueError: file lenghts is indivisible by the number of chunks requested to parallelize operations

    Returns:
        np.array: array of data from file
    """
    with open(filename, mode="rb") as file:
        filecontent = file.read()  #  serial representation
    image_dimension = 1952
    newarr = np.zeros(shape=image_dimension * image_dimension)
    chunks_n = 32
    chunk_size = len(filecontent) / chunks_n
    if chunk_size % 1 or (chunk_size / 3) % 1:
        raise ValueError("Indivisible by chunks")
    chunk_size = int(chunk_size)
    splitted = np.array(
        [filecontent[i * chunk_size : (i + 1) * chunk_size] for i in range(chunks_n)]
    )
    with mp.Pool(processes=chunks_n) as p:
        result = p.map(three_bytes_to_two_ints, splitted)
    newarr = np.array(result).reshape((1952, 1952))
    return newarr


def convert_set(filenames, new_filename, height, width):
    """ANTARTICOR ONLY: Sums a set of filenames and converts them to one fits file.


    Args:
        filenames (list): list of file names to be summed before being converted
        new_filename (str): new .fits file name
    """
    abs_new_filename = os.path.abspath(new_filename)
    if abs_new_filename.split(".")[-1] != "fits":
        raise ValueError("Trying to save a .fits file to .bin, check new filename")
    if type(filenames) is not list:
        filenames = [
            filenames,
        ]
    images_n = len(filenames)
    arr = np.zeros(shape=(height, width))
    for filename in tqdm.tqdm(filenames):
        arr += nparr_from_binary(filename) / images_n
    hdu = fits.PrimaryHDU(data=arr)
    date_and_time = datetime.datetime.now(tz=pytz.timezone("Australia/Perth")).strftime(
        "%Y-%m-%dT%H:%M:%S%z"
    )
    hdu.header["CREATED"] = (
        str(date_and_time),
        "Datetime conversion from bin to fits file (Dome C timezone).",
    )
    hdu.writeto(abs_new_filename, overwrite=True)


def average_rawfiles_to_fits(
    filenames: list, new_filename: str, height: int, width: int
):
    """Saves the mean of a list of rawfiles to a new fits file.

    Args:
        filenames (list): list of raw filenames
        new_filename (str): new fits filename
        height (int): image height in pix
        width (int): image width in pix

    Raises:
        ValueError: trying to save in a file that does not end with .fits
    """
    merge_rawfiles_to_fits(
        filenames=filenames,
        new_filename=new_filename,
        height=height,
        width=width,
        mode="average",
    )

    """
    abs_new_filename = str(Path(new_filename).absolute())
    if abs_new_filename.split(".")[-1] != "fits":
        raise ValueError("Output filename must have a .fits extension.")
    if type(filenames) is not list:
        filenames = [
            filenames,
        ]
    images_n = len(filenames)
    arr = np.zeros(shape=(height, width))
    for filename in tqdm.tqdm(filenames):
        with open(filename, mode="rb") as file:
            buffer = file.read()
        data = np.ndarray(shape=(height, width), dtype="<u2", buffer=buffer)

        arr += data / images_n
    hdu = fits.PrimaryHDU(data=arr)
    date_and_time = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")
    hdu.header["CREATED"] = (
        str(date_and_time),
        "Datetime conversion from bin to fits file (Dome C timezone).",
    )
    hdu.writeto(abs_new_filename, overwrite=True)
    """


def merge_rawfiles_to_fits(
    filenames: list, new_filename: str, height: int, width: int, mode="sum"
):
    """Saves the average or sum of a list of rawfiles to a new fits file.

    Args:
        filenames (list): list of raw filenames
        new_filename (str): new fits filename
        height (int): image height in pix
        width (int): image width in pix
        mode (str): wether to "average" or "sum" the images. Defaults to "sum".

    Raises:
        ValueError: trying to save in a file that does not end with .fits
    """
    abs_new_filename = str(Path(new_filename).absolute())
    if abs_new_filename.split(".")[-1] != "fits":
        raise ValueError("Output filename must have a .fits extension.")
    if type(filenames) is not list:
        filenames = [
            filenames,
        ]
    images_n = len(filenames)
    arr = np.zeros(shape=(height, width))
    for filename in tqdm.tqdm(filenames):
        with open(filename, mode="rb") as file:
            buffer = file.read()
        data = np.ndarray(shape=(height, width), dtype="<u2", buffer=buffer)
        arr += data

    if mode == "average":
        arr /= images_n

    hdu = fits.PrimaryHDU(data=arr)
    date_and_time = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")
    hdu.header["CREATED"] = (
        str(date_and_time),
        "Datetime conversion from bin to fits file (Dome C timezone).",
    )
    hdu.writeto(abs_new_filename, overwrite=True)


def convert_rawfile_to_fits(
    filename: str, height: int, width: int, remove_old: bool = False
):
    """Converts a raw file to a fits one, using default header

    Args:
        filename (str): raw filename
        height (int): file height
        width (int): file width
        remove_old (bool, optional): remove old raw file after conversion. Defaults to False.

    Raises:
        ValueError: raised if the file does not end with ".raw"
    """
    if not ".raw" in filename:
        raise ValueError("Can't convert: not a row file")
    with open(filename, mode="rb") as file:
        buffer = file.read()
    data = np.ndarray(shape=(height, width), dtype="<u2", buffer=buffer)
    HDU = fits.PrimaryHDU(data=data)
    filename = os.path.abspath(filename)  # prevents bug when "../" is in filename
    new_filename = os.path.pathsep.join(filename.split(".")[:-1]) + ".fits"
    HDU.writeto(new_filename, overwrite=True)
    info(f'Image successfully saved to "{new_filename}".')

    if remove_old:
        os.remove(filename)



import sys
from logging import info, warning
from pathlib import Path

import numpy as np
from astropy.io import fits
from numba import njit
from tqdm import tqdm

from micropolarray.cameras import PolarCam
from micropolarray.processing.nrgf import roi_from_polar
from micropolarray.utils import _make_abs_and_create_dir


def calculate_chen_wan_lian_calibration(
    polarizer_orientations,
    filenames_list,
    micropol_phases_previsions,
    output_dir,
    occulter=True,
    dark_filename=None,
    flat_filename=None,
):
    """Performs calibration from Chen-Wang-Liang paper 2014

    Args:
        polarizer_orientations (list[float]): List of polarizer orienataions
        filenames_list (list[str]): List of filenames coupled with
        micropol_phases_previsions (list[float]): Previsions of the micropolarizer orientations inside a superpixel
        output_dir (str): output path for the calibration matrices
        occulter (bool, optional): wether to exclude the occulter area. Defaults to True.
        dark_filename (str, optional): path to the dark to be subtracted from the images. Defaults to None.
        flat_filename (str, optional): path to the dark to be subtracted from the images. Defaults to None.

    Raises:
        ValueError: polarizer orientation list and filenames list do not have the same lenght
        ValueError: any of 0,45,90,-45 polarizarions is not included in the polarizer orientation list
    """

    DEBUG = False

    if type(filenames_list) is not list:
        filenames_list = [filenames_list]

    pols_n = len(polarizer_orientations)
    filenames_n = len(filenames_list)
    if pols_n != filenames_n:
        raise ValueError("Polarizer orientations do not match filenames.")

    if not np.all(np.isin([0, 45, 90, -45], polarizer_orientations)):
        raise ValueError(
            "All (0, 45, 90, -45) pols must be included in the polarizer orientation array"
        )
    # Have to be sorted
    polarizer_orientations, filenames_list = (
        list(t)
        for t in zip(*sorted(zip(polarizer_orientations, filenames_list)))
    )

    angle_dic = {micropol_phases_previsions[i]: i for i in range(4)}

    with fits.open(filenames_list[0]) as file:
        data = file[0].data  # get data dimension
    occulter_flag = np.ones_like(data)  # 0 if not a occulted px, 1 otherwise
    if occulter:
        # Mean values from coronal observations 2022_12_03
        # (campagna_2022/mean_occulter_pos.py)

        occulter_y, occulter_x, occulter_r = PolarCam().occulter_pos_nov2022
        occulter_r += 10  # Better to overoccult

        occulter_flag = roi_from_polar(
            occulter_flag, [occulter_y, occulter_x], [0, occulter_r]
        )
        for super_y in range(0, occulter_flag.shape[0], 2):
            for super_x in range(0, occulter_flag.shape[1], 2):
                if np.any(
                    occulter_flag[super_y : super_y + 2, super_x : super_x + 2]
                ):
                    occulter_flag[
                        super_y : super_y + 2, super_x : super_x + 2
                    ] = 1
                    continue
    else:
        occulter_flag *= 0

    # Collect dark
    if dark_filename:
        with fits.open(dark_filename) as file:
            dark = file[0].data
    # Collect flat field, normalize it
    if flat_filename:
        with fits.open(flat_filename) as file:
            flat = file[0].data
    if flat_filename and dark_filename:
        flat -= dark  # correct flat too
        flat = np.where(flat > 0, flat, 1.0)
        if occulter:
            flat = np.where(occulter_flag, 1.0, flat)

    flat_max = np.max(flat, axis=(0, 1))
    normalized_flat = np.where(occulter_flag, 1.0, flat / flat_max)

    all_data_arr = [0.0] * filenames_n
    for idx, filename in enumerate(filenames_list):
        with fits.open(filename) as file:
            all_data_arr[idx] = file[0].data
            if dark_filename is not None:
                all_data_arr[idx] -= dark
                all_data_arr[idx] = np.where(
                    all_data_arr[idx] >= 0, all_data_arr[idx], 0.0
                )
            if flat_filename is not None:
                all_data_arr[idx] = np.where(
                    normalized_flat != 0,
                    all_data_arr[idx] / normalized_flat,
                    all_data_arr[idx],
                )

    all_data_arr = np.array(all_data_arr)
    _, height, width = all_data_arr.shape

    if DEBUG:
        warning("Running in DEBUG mode")
        height = int(height / 2)
        width = int(width / 2)
        all_data_arr = all_data_arr[:, 0:height, 0:width]

    # Stokes input vector: S = [pol] * [S_0, S_1, S_2, 1] * [width, height]
    S_input = np.zeros(shape=(filenames_n, 4, height, width))

    single_pol_images = np.array(
        [
            [
                all_data_arr[pol_idx, y::2, x::2]
                for y in range(2)
                for x in range(2)
            ]
            for pol_idx in range(pols_n)
        ]
    )

    sorted_single_pol_images = np.array(single_pol_images, copy=True)
    sorted_single_pol_images[:, 0, :, :] = single_pol_images[
        :, angle_dic[0], :, :
    ]
    sorted_single_pol_images[:, 1, :, :] = single_pol_images[
        :, angle_dic[45], :, :
    ]
    sorted_single_pol_images[:, 2, :, :] = single_pol_images[
        :, angle_dic[-45], :, :
    ]
    sorted_single_pol_images[:, 3, :, :] = single_pol_images[
        :, angle_dic[90], :, :
    ]

    _build_S_input(sorted_single_pol_images, S_input)  # speeds up A LOT

    W_ideal = np.array(
        [
            [0.5, 0.5, 0.5, 0.5],
            [
                0.5 * np.cos(2.0 * np.deg2rad(micropol_phases_previsions[i]))
                for i in range(4)
            ],
            [
                0.5 * np.sin(2.0 * np.deg2rad(micropol_phases_previsions[i]))
                for i in range(4)
            ],
        ]
    ).T  # shape=(4,3)

    C_ij = np.zeros(shape=(pols_n, 4, 4, height, width))  # correction matrix
    d_ij = np.zeros(
        shape=(pols_n, height, width)
    )  # offsets, each element is equal in the superpixel so we can save one image with 4 pixel to save space

    if DEBUG:
        height = 8
        width = 8

    for pol_idx in tqdm(range(pols_n)):
        for super_y in tqdm(range(0, height, 2)):
            for super_x in range(0, width, 2):
                S = np.array([S_input[pol_idx, :, super_y, super_x]]).T

                S_inv_ij = np.linalg.pinv(S)
                P_ij = all_data_arr[
                    pol_idx, super_y : super_y + 2, super_x : super_x + 2
                ].reshape(4, 1)

                W_d = P_ij @ S_inv_ij
                # W_d = W_d.T
                W = W_d[:, :3]
                d = W_d[:, 3]
                try:
                    W_inv = np.linalg.pinv(W)
                except np.linalg.LinAlgError:
                    W_inv = 0.0 * W.T

                d_ij[
                    pol_idx, super_y : super_y + 2, super_x : super_x + 2
                ] = d.reshape(2, 2)

                C = W_ideal @ W_inv
                for i in range(2):
                    for j in range(2):
                        C_ij[pol_idx, :, :, super_y + i, super_x + j] = C

    C_ij = np.mean(C_ij, axis=0)
    d_ij = np.mean(d_ij, axis=0)

    output_dir = _make_abs_and_create_dir(output_dir)
    for i in range(4):
        for j in range(4):
            hdu = fits.PrimaryHDU(data=C_ij[i, j])
            hdu.writeto(output_dir + f"/C{i}{j}.fits", overwrite=True)
    hdu = fits.PrimaryHDU(data=d_ij)
    hdu.writeto(output_dir + "/d_ij.fits", overwrite=True)


@njit
def _build_S_input(single_pol_images, S_input):
    # ANGLES_LIST positions:
    #    0 <==> 0
    #    1 <==> 45
    #    2 <==> -45
    #    3 <==> 90
    pols_n = single_pol_images.shape[0]
    height = single_pol_images.shape[2] * 2
    width = single_pol_images.shape[3] * 2

    S_input[:, -1, :, :] += 1
    # Initialize stokes vector
    for pol_idx in range(pols_n):
        for super_y in range(0, height, 2):
            y = int(super_y / 2)
            for super_x in range(0, width, 2):
                x = int(super_x / 2)
                S_input[
                    pol_idx, 0, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 0, y, x]
                    + single_pol_images[pol_idx, 3, y, x]
                )
                S_input[
                    pol_idx, 1, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 0, y, x]
                    - single_pol_images[pol_idx, 3, y, x]
                )
                S_input[
                    pol_idx, 2, super_y : super_y + 2, super_x : super_x + 2
                ] = (
                    single_pol_images[pol_idx, 1, y, x]
                    - single_pol_images[pol_idx, 2, y, x]
                )


def chen_wan_liang_calibration(data, calibration_matrices_dir: str):
    """Calibrates the images using Chen-Wang-Liang 2014 paper calibration

    Args:
        data (np.array): data to be calibrated
        calibration_matrices_dir (str): path to the calibration matrices

    Returns:
        np.array: calibrated data
    """
    info("Applying Chen-Wan-Liang correction...")
    height, width = data.shape
    C_ij = np.zeros(shape=(4, 4, height, width))
    d_ij = np.zeros_like(data)
    corrected_data = np.zeros_like(data)
    ifov_corrected_data = np.zeros_like(data)
    correction_path = Path(calibration_matrices_dir)
    if not correction_path.is_absolute():  # suppose it is in cwd
        correction_path = correction_path.joinpath(
            Path().cwd(), correction_path
        )

    for i in range(4):
        for j in range(4):
            with fits.open(
                calibration_matrices_dir + f"/C{i}{j}.fits"
            ) as file:
                C_ij[i, j] = file[0].data
    with fits.open(calibration_matrices_dir + f"/d_ij.fits") as file:
        d_ij = file[0].data

    P_minus_d = data - d_ij

    for super_y in range(0, height, 2):
        for super_x in range(0, width, 2):
            mult = C_ij[:, :, super_y, super_x] @ P_minus_d[
                super_y : super_y + 2, super_x : super_x + 2
            ].reshape(4, 1)

            corrected_data[
                super_y : super_y + 2, super_x : super_x + 2
            ] = mult.reshape(2, 2)

    ifov_corrected_data = _ifov_jitcorrect(corrected_data, height, width)

    return ifov_corrected_data


@njit
def _ifov_jitcorrect(data, height, width):
    corrected_data = np.zeros(shape=(height, width))
    # correct IFOV error, skip first rows/cols
    for super_y in range(2, height, 2):
        for super_x in range(2, width, 2):
            for i in range(2):
                for j in range(2):
                    corrected_data[super_y + j, super_x + i] = 0.25 * (
                        data[super_y + j - 2, super_x + i - 2]
                        + data[super_y + j - 2, super_x + i]
                        + data[super_y + j, super_x + i - 2]
                        + data[super_y + j, super_x + i]
                    )
    return corrected_data



import glob
import os
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.io import fits
from scipy.optimize import curve_fit
from test_utils import generate_dummy_data, generate_polarized_data

import micropolarray as ml
from micropolarray.polarization_functions import AoLP, DoLP, pB
from micropolarray.processing.demodulation import Malus


class TestMicropolImage:
    def write_temp_image(self, tmp_path, data):
        """Writes images to the temp folder for testing"""
        image = fits.PrimaryHDU(
            data=data, do_not_scale_image_data=True, uint=False
        )
        image.header["FROMFILE"] = True
        image.writeto(tmp_path / "sample_image.fits")

    def test_image_initialization(self, generate_dummy_data, tmp_path):
        """Tests the initialization of both Image and MicroPolArrayImage"""
        dummy_data_16 = generate_dummy_data(16)
        self.write_temp_image(tmp_path, dummy_data_16)
        for ImageClass in [ml.Image, ml.MicropolImage]:
            image = ImageClass(dummy_data_16)
            assert np.all(image.data == dummy_data_16)

            image = ImageClass(str(tmp_path / "sample_image.fits"))
            assert np.all(image.data == dummy_data_16)
            assert image.header["FROMFILE"] == True

            image = ImageClass(image)
            assert np.all(image.data == dummy_data_16)

    def test_image_writing(self, generate_dummy_data, tmp_path):
        """Tests the saving of both Image and MicroPolArrayImage"""
        dummy_data_16 = generate_dummy_data(16)
        image = ml.Image(dummy_data_16)
        for input_path in ["image.fits", "test/image.fits"]:
            image.save_as_fits(tmp_path / input_path)
            read_image = ml.Image(
                str((tmp_path / input_path).with_suffix(".fits"))
            )
            assert np.all(read_image.data == dummy_data_16)
            read_image.header["EXTEND"] = True
            assert read_image.header == image.header

        for image_type in [ml.Image, ml.MicropolImage]:
            image = image_type(dummy_data_16)
            image.save_as_fits(str(tmp_path / "image.fits"))

        image = ml.MicroPolarizerArrayImage(dummy_data_16)
        print(tmp_path / "image.fits")
        print((tmp_path / "image.fits").suffix)

        image.save_single_pol_images(tmp_path / "image.fits")

        for i in glob.glob(str(tmp_path / "image*")):
            print(i)
        return
        assert np.all(ml.Image(tmp_path / "image_POL0.fits").data == 1)

    def test_show(self, generate_dummy_data):
        dummy_data_16 = generate_dummy_data(16)
        for image_type in [ml.MicropolImage, ml.Image]:
            dummy_image = image_type(dummy_data_16)
            dummy_image.show()
            dummy_image.show_histogram()
        dummy_image = ml.MicropolImage(dummy_data_16)
        dummy_image.show_with_pol_params()
        dummy_image.show_single_pol_images()
        dummy_image.show_pol_param("DoLP")
        dummy_image.demosaic()
        dummy_image.show_demo_images()

    def test_dark_and_flat_correction(self, generate_dummy_data, tmp_path):
        # test dark
        dummy_data_16 = generate_dummy_data(16)
        dark_data = generate_dummy_data(16)
        dark_image = ml.MicropolImage(dark_data)
        dummy_image = ml.MicropolImage(dummy_data_16, dark=dark_image)
        assert np.all(dummy_image.data == 0.0)
        assert np.all(dummy_image.DoLP.data == 0.0)
        # test flat
        signal = 4.0
        dummy_data_16 = np.ones(shape=(16, 16)) * signal
        flat_image = ml.MicropolImage(dummy_data_16 * np.random.random(1))
        dummy_image = ml.MicropolImage(dummy_data_16, flat=flat_image)
        assert np.all(dummy_image.data == signal)

    def test_demosaic(self, generate_dummy_data, tmp_path):
        """Tests demosaic operation and demosaic writing"""
        dummy_data_16 = generate_dummy_data(16)
        # test mean
        image = ml.MicropolImage(dummy_data_16)
        assert image.data.shape == (16, 16)
        assert image.I.data.shape == (8, 8)
        image = image.demosaic(demosaic_mode="mean")

        for idx, demo_image in enumerate(image.demosaiced_images):
            assert np.all(demo_image == np.full((16, 16), (idx + 1) / 4.0))
        assert np.all(
            image.I.data == np.full((16, 16), 0.25 * 0.5 * (1 + 2 + 3 + 4))
        )

        # test adjacent
        image = ml.MicropolImage(dummy_data_16)
        image = image.demosaic(demosaic_mode="adjacent")
        for idx, demo_image in enumerate(image.demosaiced_images):
            assert np.all(demo_image == np.full((16, 16), (idx + 1)))

        # test writing
        image.save_demosaiced_images_as_fits(
            str(tmp_path / "demosaiced_images.fits")
        )

    def test_rebinning(self, generate_dummy_data):
        """Tests 2x2 and 4x4 binning (the other will be supposedly fine)"""
        dummy_data_16 = generate_dummy_data(16)

        binned_image_2 = ml.MicropolImage(dummy_data_16).rebin(2)
        assert np.all(binned_image_2.data == generate_dummy_data(8) * 4)

        binned_image_4 = ml.MicropolImage(dummy_data_16).rebin(4)
        assert np.all(binned_image_4.data == generate_dummy_data(4) * 16)

    def test_pol_parameters(self, generate_dummy_data):
        """Test if polarization parameters are correcly computed"""

        def test_theo_stokes(image, I, Q, U):
            assert np.all(image.I.data == I)
            assert np.all(image.Q.data == Q)
            assert np.all(image.U.data == U)

            assert np.all(
                image.AoLP.data == 0.5 * np.arctan2(U, Q) * half_ones
            )
            assert np.all(image.pB.data == np.sqrt(Q * Q + U * U) * half_ones)
            assert np.all(
                image.DoLP.data == np.sqrt(Q * Q + U * U) * half_ones / I
            )

        array_side = 16
        dummy_data_16 = generate_dummy_data(array_side)
        half_ones = np.ones(shape=(int(array_side / 2), int(array_side / 2)))

        image = ml.MicropolImage(dummy_data_16)

        for i in range(4):
            assert np.all(image.single_pol_subimages[i] == i + 1)

        angles = [0, 45, -45, 90]
        numbers = [1.0, 2.0, 3.0, 4.0]
        for angle, n in zip(angles, numbers):
            assert np.all(
                image.single_pol_subimages[image.angle_dic[angle]] == n
            )

        I = 0.5 * (1 + 2 + 3 + 4)
        Q = 1.0 - 4.0
        U = 2.0 - 3.0
        test_theo_stokes(image, I, Q, U)

        new_angle_dic = {45: 0, 0: 1, 90: 2, -45: 3}
        image = ml.MicropolImage(dummy_data_16, angle_dic=new_angle_dic)
        assert image.angle_dic == new_angle_dic

        Q = 2.0 - 3.0
        U = 1.0 - 4.0
        test_theo_stokes(image, I, Q, U)

    def test_congrid(self, generate_dummy_data):
        dummy_data_40 = generate_dummy_data(40)
        dummy_data_16 = generate_dummy_data(16)
        image = ml.MicropolImage(dummy_data_40)
        congridded_image = image.congrid(16, 16)
        assert np.all(congridded_image.data == dummy_data_16)
        assert np.all(congridded_image.I.data == 0.5 * (1 + 2 + 3 + 4))

    def test_operations(self, generate_dummy_data):
        data = np.ones(shape=(16, 16))
        for image_type in [ml.Image, ml.MicropolImage]:
            image1 = image_type(7 * data)
            image2 = image_type(2.0 * data)
            result = image1 + image2
            assert np.all(result.data == 9.0)
            result = image1 - image2
            assert np.all(result.data == 5.0)
            result = image1 * image2
            assert np.all(result.data == 14.0)
            result = image1 / image2
            assert np.all(result.data == 3.5)

        image1 = ml.MicropolImage(
            generate_polarized_data(shape=(6, 16), S=100, angle_rad=0)
        )
        image2 = ml.MicropolImage(50 * np.ones_like(image1.data))

        print(image1.AoLP)



import glob
import os
from logging import info, warn
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pytest
from astropy.io import fits
from scipy.optimize import curve_fit
from test_utils import generate_dummy_data, generate_polarized_data

import micropolarray as ml
from micropolarray.polarization_functions import AoLP, DoLP, pB
from micropolarray.processing.demodulation import Malus


class TestDemodulation:
    def test_demo_from_dummy(self, generate_dummy_data, tmp_path):
        """Create a dummy demodulation matrix, save it, read it then use it to demodulate. Check if demodulation is correctly done."""
        ml.set_default_angles(ml.PolarCam().angle_dic)
        angles = np.array([np.deg2rad(angle) for angle in [0, 45, -45, 90]])
        demo_matrix = np.array(
            [
                [0.5, 0.5, 0.5, 0.5],
                np.cos(2 * angles),
                np.sin(2 * angles),
            ]
        )
        for i in range(3):
            for j in range(4):
                image = fits.PrimaryHDU()
                image.data = np.ones(shape=(8, 8)) * demo_matrix[i, j]
                image.writeto(tmp_path / f"M{i}{j}.fits")

        image = ml.PolarcamImage(generate_dummy_data(16))
        assert np.all(image.Q.data == (1 - 4))
        assert np.all(image.U.data == (2 - 3))
        assert np.all(image.I.data == (0.5 * (1 + 2 + 3 + 4)))

        demodulator = ml.Demodulator(str(tmp_path))

        demo_image = image.demodulate(demodulator=demodulator)
        assert np.all(np.round(demo_image.Q.data, 5) == (1.0 - 4.0))
        assert np.all(np.round(demo_image.U.data, 5) == (2.0 - 3.0))
        assert np.all(np.round(demo_image.I.data, 5) == (0.5 * (1.0 + 2.0 + 3.0 + 4.0)))

    # TODO refactor calling angles with fixtures
    def test_demodulation_computation(self, tmp_path):
        # try demodulation
        angles = np.array([np.deg2rad(angle) for angle in [0, 45, -45, 90]])
        output_dir = tmp_path / "computed_matrix"
        output_str = str(output_dir)

        polarizations = np.arange(-45, 136, 15)
        pols_rad = np.deg2rad(polarizations)
        input_signal = 100
        t = 0.9
        eff = 0.7
        side = 20
        shape = (side, side)
        ones = np.ones(shape=shape)
        for pol, pol_rad in zip(polarizations, pols_rad):
            result_image = ml.MicropolImage(
                generate_polarized_data(
                    shape=shape,
                    S=input_signal,
                    angle_rad=pol_rad,
                    t=t,
                    eff=eff,
                )
            )
            result_image.save_as_fits(tmp_path / f"pol_{int(pol)}.fits")
        if False:  # check that fit will be ok
            for angle in angles:
                pars, pcov = curve_fit(
                    Malus,
                    pols_rad,
                    np.array([Malus(pol, t, eff, angle) for pol in pols_rad]),
                )
                print(f"t = {pars[0]}")
                print(f"eff = {pars[1] }")
                print(f"phi = {np.rad2deg(pars[2]) }")

        # read the files
        filenames = sorted(
            glob.glob(str(tmp_path / "pol*.fits")),
            key=lambda x: int(x.split(os.path.sep)[-1][4:].strip(".fits")),
        )
        # test normalization with number and with array
        for S_type in [
            input_signal,
            input_signal * np.ones(shape=(side, side)),
        ]:
            ml.calculate_demodulation_tensor(
                polarizer_orientations=polarizations,
                filenames_list=filenames,
                micropol_phases_previsions=np.rad2deg(angles),
                gain=2.75,
                output_dir=output_str,
                binning=1,
                occulter=False,
                procs_grid=[2, 2],
                dark_filename=None,
                flat_filename=None,
                normalizing_S=S_type,
                DEBUG=False,
            )

        # image polarized with phi=0 uniform, t=1, eff=1
        ideal_image = ml.MicropolImage(
            generate_polarized_data(
                shape=shape, S=input_signal, angle_rad=0, t=1, eff=1
            )
        )
        assert np.all(ideal_image.I.data == input_signal)
        assert np.all(ideal_image.Q.data == input_signal)
        assert np.all(ideal_image.U.data == 0)
        assert np.all(ideal_image.pB.data == input_signal)
        assert np.all(ideal_image.AoLP.data == 0)
        assert np.all(ideal_image.DoLP.data == 1)

        demodulator = ml.Demodulator(output_str)

        assert demodulator.eff.shape == ones.shape
        assert np.all(np.round(demodulator.eff, 2) == np.round(eff * ones, 2))
        assert np.all(np.round(demodulator.tk, 2) == np.round(t * ones, 2))
        assert np.all(
            np.round(demodulator.phi, 2)
            == np.round(
                ml.merge_polarizations(
                    [
                        np.rad2deg(angles[i])
                        * np.ones(shape=(int(side / 2), int(side / 2)))
                        for i in range(4)
                    ]
                ),
                2,
            )
        )

        assert (
            demodulator.fit_found_flags.shape == demodulator.mij[0, 0].shape
        )  # after September 2023
        test_angle = np.deg2rad(30)
        example_image = ml.MicropolImage(
            generate_polarized_data(
                shape, S=input_signal, angle_rad=test_angle, t=t, eff=eff
            )
        )
        # Theoric values
        I = input_signal * (
            Malus(test_angle, 1, 1, 0) + Malus(test_angle, 1, 1, np.pi / 2)
        )
        Q = input_signal * (
            Malus(test_angle, 1, 1, 0) - Malus(test_angle, 1, 1, np.pi / 2)
        )
        U = input_signal * (
            Malus(test_angle, 1, 1, np.pi / 4) - Malus(test_angle, 1, 1, -np.pi / 4)
        )
        S = [I, Q, U]
        dolp = np.round(DoLP(S), 5)
        aolp = np.round(AoLP(S), 5)
        pb = np.round(pB(S), 5)

        # test demodulation without demosaicing
        no_demosaic_image = ml.MicropolImage(example_image)
        no_demosaic_image = no_demosaic_image.demodulate(demodulator, demosaicing=False)
        assert np.all(np.round(no_demosaic_image.pB.data, 5) == pb)
        assert (
            no_demosaic_image.pB.data.shape
            == no_demosaic_image.single_pol_subimages[0].shape
        )

        demo_image = example_image.demodulate(demodulator)
        if False:
            demodulator.show()
            example_image.show_with_pol_params()
            plt.show()

        assert np.all(np.round(demo_image.I.data, 5) == np.round(I, 5))
        assert np.all(np.round(demo_image.Q.data, 5) == np.round(Q, 5))
        assert np.all(np.round(demo_image.U.data, 5) == np.round(U, 5))
        assert np.all(np.round(demo_image.DoLP.data, 5) == dolp)
        assert np.all(np.round(demo_image.AoLP.data, 5) == aolp)
        assert np.all(np.round(demo_image.pB.data, 5) == pb)

        simples = []
        measureds = []
        theos = []

        def normalize(angle):
            while angle > np.pi / 2:
                angle -= np.pi
            while angle < -np.pi / 2:
                angle += np.pi
            return angle

        # for dummy_angle in np.arange(-np.pi / 2, np.pi / 2, 0.1):
        for dummy_angle in np.arange(-np.pi, np.pi, 0.1):
            polarized_image = ml.MicropolImage(
                generate_polarized_data(shape, input_signal, dummy_angle, t, eff)
            )

            simple = np.round(np.mean(polarized_image.AoLP.data), 1)
            simples.append(simple)

            polarized_image = polarized_image.demodulate(demodulator)

            measured = np.round(np.mean(polarized_image.AoLP.data), 1)
            theo = np.round(normalize(dummy_angle), 1)
            measureds.append(measured)
            theos.append(theo)

        # debug
        if False:
            for measured, theo, simple in zip(measureds, theos, simples):
                print(theo, simple, measured)
            fig, ax = plt.subplots()
            x = np.arange(-np.pi / 2, np.pi / 2, 0.1)
            x = np.arange(-np.pi, np.pi, 0.1)
            ax.plot(x, theos, label="theos")
            ax.plot(x, simples, label="simple")
            ax.plot(x, measureds, label="demodulated")
            ax.legend()
            plt.show()

        assert np.all(simple == theo)
        assert np.all(measured == theo)

    def test_demo_rebin(self, generate_dummy_data, tmp_path):
        dummy_data = generate_dummy_data(32)
